import os
import configparser
from functools import cached_property


class FigConfig(configparser.SafeConfigParser):
    ALL_BACKENDS = {'AWS', 'AWS_SQS', 'AZURE', 'GCP', 'WORKSPACEONE', 'CHRONICLE', 'CLOUDTRAIL_LAKE'}
    FALCON_CLOUD_REGIONS = {'us-1', 'us-2', 'eu-1', 'us-gov-1'}
    SENSOR_RECOGNIZED_CLOUDS = {'AWS', 'Azure', 'GCP', 'unrecognized'}
    ENV_DEFAULTS = [
        ['main', 'backends', 'FIG_BACKENDS'],
        ['main', 'worker_threads', 'FIG_WORKER_THREADS'],
        ['logging', 'level', 'LOG_LEVEL'],
        ['events', 'severity_threshold', 'EVENTS_SEVERITY_THRESHOLD'],
        ['events', 'older_than_days_threshold', 'EVENTS_OLDER_THAN_DAYS_THRESHOLD'],
        ['falcon', 'cloud_region', 'FALCON_CLOUD_REGION'],
        ['falcon', 'client_id', 'FALCON_CLIENT_ID'],
        ['falcon', 'client_secret', 'FALCON_CLIENT_SECRET'],
        ['falcon', 'reconnect_retry_count', 'FALCON_RECONNECT_RETRY_COUNT'],
        ['falcon', 'application_id', 'FALCON_APPLICATION_ID'],
        ['azure', 'workspace_id', 'WORKSPACE_ID'],
        ['azure', 'primary_key', 'PRIMARY_KEY'],
        ['aws', 'region', 'AWS_REGION'],
        ['aws_sqs', 'region', 'AWS_REGION'],
        ['aws_sqs', 'sqs_queue_name', 'AWS_SQS'],
        ['workspaceone', 'token', 'WORKSPACEONE_TOKEN'],
        ['workspaceone', 'syslog_host', 'SYSLOG_HOST'],
        ['workspaceone', 'syslog_port', 'SYSLOG_PORT'],
        ['chronicle', 'service_account', 'GOOGLE_SERVICE_ACCOUNT_FILE'],
        ['chronicle', 'customer_id', 'GOOGLE_CUSTOMER_ID'],
        ['chronicle', 'region', 'CHRONICLE_REGION'],
        ['cloudtrail_lake', 'channel_arn', 'CLOUDTRAIL_LAKE_CHANNEL_ARN'],
        ['cloudtrail_lake', 'region', 'CLOUDTRAIL_LAKE_REGION'],
    ]

    def __init__(self):
        super().__init__()
        self.read(['config/defaults.ini', 'config/config.ini', 'config/devel.ini'])
        self._override_from_env()

    def _override_from_env(self):
        for section, var, envvar in self.__class__.ENV_DEFAULTS:
            value = os.getenv(envvar)
            if value:
                self.set(section, var, value)

    def validate(self):
        for section, var, envvar in self.__class__.ENV_DEFAULTS:
            try:
                self.get(section, var)
            except configparser.NoOptionError as err:
                raise Exception(
                    "Please provide environment variable {} or configuration option {}.{}".format(
                        envvar, section, var)) from err

        if int(self.get('main', 'worker_threads')) not in range(1, 128):
            raise Exception('Malformed configuration: expected main.worker_threads to be in range 1-128')
        self.validate_falcon()
        self.validate_events()
        self.validate_backends()

    def validate_falcon(self):
        if int(self.get('falcon', 'reconnect_retry_count')) not in range(1, 10000):
            raise Exception('Malformed configuration: expected falcon.reconnect_retry_count to be in range 0-10000')
        if self.get('falcon', 'cloud_region') not in self.FALCON_CLOUD_REGIONS:
            raise Exception(
                'Malformed configuration: expected falcon.cloud_region to be in {}'.format(self.FALCON_CLOUD_REGIONS)
            )

    def validate_events(self):
        if not self.detections_exclude_clouds.issubset(self.SENSOR_RECOGNIZED_CLOUDS):
            raise Exception(
                'Malformed configuration: expected events.detections_exclude_clouds to be subset of "{}" got "{}"'.format(
                    self.SENSOR_RECOGNIZED_CLOUDS, self.detections_exclude_clouds))

        if int(self.get('events', 'severity_threshold')) not in range(0, 5):
            raise Exception('Malformed configuration: expected events.severity_threshold to be in range 0-4')
        if int(self.get('events', 'older_than_days_threshold')) not in range(0, 10000):
            raise Exception('Malformed configuration: expected events.older_than_days_threshold to be in range 0-10000')

    def validate_backends(self):
        if not self.backends.issubset(self.ALL_BACKENDS) or len(self.backends) < 1:
            raise Exception(
                'Malformed configuration: expected main.backends to be subset of "{}" and contain at least one'.format(
                    self.ALL_BACKENDS))
        if 'AWS' in self.backends:
            if len(self.get('aws', 'region')) == 0:
                raise Exception('Malformed Configuration: expected aws.region to be non-empty')
        if 'AWS_SQS' in self.backends:
            if len(self.get('aws_sqs', 'region')) == 0:
                raise Exception('Malformed Configuration: expected aws_sqs.region to be non-empty')
            if len(self.get('aws_sqs', 'sqs_queue_name')) == 0:
                raise Exception('Malformed Configuration: expected aws_sqs.sqs_queue_name to be non-empty')
        if 'WORKSPACEONE' in self.backends:
            if len(self.get('workspaceone', 'token')) == 0:
                raise Exception('Malformed Configuration: expected workspaceone.token to be non-empty')
            if len(self.get('workspaceone', 'syslog_host')) == 0:
                raise Exception('Malformed Configuration: expected workspaceone.syslog_host to be non-empty')
            if int(self.get('workspaceone', 'syslog_port')) not in range(1, 65535):
                raise Exception('Malformed configuration: expected workspaceone.syslog_port to be in range 1-65335')
        if 'CHRONICLE' in self.backends:
            if len(self.get('chronicle', 'service_account')) == 0:
                raise Exception('Malformed Configuration: expected chronicle.service_account_file to be non-empty')
            if len(self.get('chronicle', 'region')) == 0:
                raise Exception('Malformed Configuration: expected chronicle.region to be non-empty')
            if len(self.get('chronicle', 'customer_id')) == 0:
                raise Exception('Malformed Configuration: expected chronicle.customer_id to be non-empty')
        if 'CLOUDTRAIL_LAKE' in self.backends:
            if len(self.get('cloudtrail_lake', 'channel_arn')) == 0:
                raise Exception('Malformed Configuration: expected cloudtrail_lake.channel_arn to be non-empty')
            if len(self.get('cloudtrail_lake', 'region')) == 0:
                raise Exception('Malformed Configuration: expected cloudtrail_lake.region to be non-empty')
        if 'AZURE' in self.backends:
            if len(self.get('azure', 'workspace_id')) == 0:
                raise Exception('Malformed Configuration: expected azure.workspace_id to be non-empty')
            if len(self.get('azure', 'primary_key')) == 0:
                raise Exception('Malformed Configuration: expected azure.primary_key to be non-empty')
            if self.get('azure', 'arc_autodiscovery') not in ['false', 'true']:
                raise Exception('Malformed Configuration: expected azure.arc_autodiscovery must be either true or false')

    @cached_property
    def backends(self):
        return set(self.get('main', 'backends').split(','))

    @cached_property
    def detections_exclude_clouds(self):
        value = self.get('events', 'detections_exclude_clouds')
        if value == '':
            return set()
        return set(value.split(','))


config = FigConfig()



import ast
import threading
import boto3
from botocore.exceptions import ClientError
from ...config import config
from ...log import log


_offset_lock = threading.RLock()


class LastEventOffset():
    '''
    Manage the state of the last seen offset for each feed with SSM paramater store.
    This is used to determine where to start the next query for each feed upon init/restart
    to prevent duplicate events.
    '''
    def __init__(self):
        self.last_seen_offsets = {}
        self.param_name = 'last_seen_offsets'
        self.client = boto3.client('ssm', region_name=config.get('cloudtrail_lake', 'region'))
        self.validate_ssm_parameter()
        self.cache = None

    def validate_ssm_parameter(self):
        '''
        Ensure the SSM parameter exists.
        '''
        try:
            self.client.get_parameter(Name=self.param_name)
        except ClientError as err:
            if err.response['Error']['Code'] == 'ParameterNotFound':
                log.info("SSM parameter %s does not exist. Creating...", self.param_name)
                self.client.put_parameter(
                    Name=self.param_name,
                    Value='{}',
                    Type='String',
                    Overwrite=True
                )
            else:
                raise err

    def get_last_seen_offsets(self):
        '''
        Get the last seen offset for all feeds via local cache or SSM parameter store.
        '''
        with _offset_lock:
            if self.cache is None:
                self.cache = self._get_remote()
            return self.cache

    def update_last_seen_offsets(self, feed_id, offset):
        '''
        Update the last seen offset for a given feed.
        '''
        with _offset_lock:
            if self.get_last_seen_offsets().get(feed_id, 0) < offset:
                self.cache[feed_id] = offset
                self._put_remote(self.cache)
                log.info("Updated last seen offset for stream feed: %s to: %s", feed_id, offset)

    def _put_remote(self, value):
        '''
        Put a value to the SSM parameter store.
        '''
        try:
            self.client.put_parameter(
                Name=self.param_name,
                Value=str(value),
                Type='String',
                Overwrite=True
            )
        except ClientError as err:
            log.exception("Failed to put value to SSM parameter with error: %s", str(err))

    def _get_remote(self):
        '''
        Get a value from the SSM parameter store.
        '''
        try:
            response = self.client.get_parameter(Name=self.param_name)
            return ast.literal_eval(response['Parameter']['Value'])
        except ClientError as err:
            log.exception("Failed to get value from SSM parameter with error: %s", str(err))
            return {}



class ApiError(Exception):
    pass


class NoStreamsError(ApiError):
    def __init__(self, app_id):
        super().__init__(
            'Falcon Streaming API not discovered. This may be caused by second instance of this application '
            'already running in your environment with the same application_id={}, or by missing streaming API '
            'capability.'.format(app_id))


class RTRError(ApiError):
    pass


class RTRConnectionError(ApiError):
    pass



import threading
from .log import log
from .falcon_data import FalconEvent, EventDataError


class WorkerThread(threading.Thread):
    def __init__(self, input_queue, falcon_cache, backends, *args, **kwargs):
        kwargs['name'] = kwargs.get('name', 'worker')
        super().__init__(*args, **kwargs)
        self.input_queue = input_queue
        self.cache = falcon_cache
        self.backends = backends

    def run(self):
        while True:
            try:
                event = self.input_queue.get()
                self.process_event(event)
            except EventDataError:
                log.exception("Could not translate falcon event to cloud provider")
            except Exception:  # pylint: disable=W0703
                log.exception("Error occurred while processing event %s", event)

    def process_event(self, event):
        falcon_event = FalconEvent(event, self.cache)
        self.backends.process(falcon_event)



from falconpy import api_complete as FalconSDK
from ..config import config
from .errors import ApiError, NoStreamsError
from .models import Stream
from .rtr import RTRSession
from .. import __version__


class FalconAPI():
    CLOUD_REGIONS = {
        'us-1': 'api.crowdstrike.com',
        'us-2': 'api.us-2.crowdstrike.com',
        'eu-1': 'api.eu-1.crowdstrike.com',
        'us-gov-1': 'api.laggar.gcw.crowdstrike.com',
    }

    def __init__(self):
        self.client = FalconSDK.APIHarness(creds={
            'client_id': config.get('falcon', 'client_id'),
            'client_secret': config.get('falcon', 'client_secret')},
            user_agent=f"falcon-integration-gateway/{__version__}",
            base_url=self.__class__.base_url())

    @classmethod
    def base_url(cls):
        return 'https://' + cls.CLOUD_REGIONS[config.get('falcon', 'cloud_region')]

    def streams(self, app_id):
        resources = self._resources(action='listAvailableStreamsOAuth2',
                                    parameters={'appId': config.get('falcon', 'application_id')})
        if not resources:
            raise NoStreamsError(app_id)
        return (Stream(s) for s in resources)

    def refresh_streaming_session(self, app_id, stream):
        self._command(action='refreshActiveStreamSession',
                      partition=stream.partition,
                      parameters={
                          'action_name': 'refresh_active_stream_session',
                          'appId': app_id
                      })

    def device_details(self, device_id):
        return self._resources(action='GetDeviceDetailsV2', ids=[device_id])

    def init_rtr_session(self, device_id):
        return self._resources(
            action='RTR_InitSession',
            body={
                'device_id': device_id
            }
        )

    def execute_rtr_command(self, action, session_id, base_command, command_string):
        return self._resources(
            action=action,
            body={
                'base_command': base_command,
                'command_string': command_string,
                'session_id': session_id
            }
        )

    def check_rtr_command_status(self, cloud_request_id, sequence_id):
        return self._resources(
            action='RTR_CheckCommandStatus',
            parameters={
                'cloud_request_id': cloud_request_id,
                'sequence_id': sequence_id,
            }
        )

    def rtr_fetch_file(self, device_id, filepath):
        session = RTRSession(self, device_id)

        z7pack = None
        try:
            z7pack = session.get_file(filepath)
        finally:
            session.close()

        import io  # pylint: disable=C0415
        import py7zr  # pylint: disable=C0415

        flo = io.BytesIO(z7pack)
        with py7zr.SevenZipFile(flo, password=config.get('falcon', 'rtr_quarantine_keyword')) as archive:
            content = archive.readall()
            if len(content) != 1:
                raise ApiError('Cannot extract RTR file from 7z')

            for _fname, bio in content.items():
                return bio.read()
        raise ApiError(f'Cannot extract file {filepath} from device {device_id}')

    def _resources(self, *args, **kwargs):
        response = self._command(*args, **kwargs)
        body = response['body']
        if 'resources' not in body or not body['resources']:
            return []
        return body['resources']

    def _command(self, *args, **kwargs):
        response = self.client.command(*args, **kwargs)
        body = response['body']
        trace_id = body.get('meta', {}).get('trace_id', '')
        if 'errors' in body and body['errors'] is not None:
            if len(body['errors']) > 0:
                raise ApiError('Error received from CrowdStrike Falcon platform: {} (trace_id: {})'.format(body['errors'], trace_id))
        if 'status_code' not in response or (response['status_code'] != 200 and response['status_code'] != 201):
            raise ApiError('Unexpected response code from Falcon API. Response was: {} (trace_id: {})'.format(response, trace_id))
        return response



from .falcon import FalconAPI, StreamManagementThread
from .worker import WorkerThread
from .queue import falcon_events
from .config import config
from .backends import Backends
from .falcon_data import FalconCache
from .log import log
from . import __version__


if __name__ == "__main__":
    log.info("Starting Falcon Integration Gateway %s", __version__)

    # Central to the fig architecture is a message queue (falcon_events). GCPWorkerThread/s read the queue and process
    # each event. The events are put on queue by StreamingThread. StreamingThread is restarted by StreamManagementThread

    config.validate()

    falcon_cache = FalconCache(FalconAPI())
    backends = Backends()

    StreamManagementThread(output_queue=falcon_events, relevant_event_types=backends.relevant_event_types).start()

    for i in range(int(config.get('main', 'worker_threads'))):
        WorkerThread(name='worker-' + str(i),
                     input_queue=falcon_events,
                     falcon_cache=falcon_cache,
                     backends=backends,
                     daemon=True).start()



import json
from .falcon import Event
from .log import log


class TranslatorError(Exception):
    pass


class EventDataError(TranslatorError):
    pass


class FalconAPIDataError(TranslatorError):
    pass


class FalconCache():
    def __init__(self, falcon_api):
        self.falcon_api = falcon_api
        self._host_detail = {}
        self._mdm_id = {}
        self._arc_config = {}

    def device_details(self, sensor_id):
        if not sensor_id:
            return EventDataError("Cannot process event. SensorId field is missing: ")

        if sensor_id not in self._host_detail:
            resources = self.falcon_api.device_details(sensor_id)
            if len(resources) > 1:
                raise FalconAPIDataError(
                    'Cannot process event for device: {}, multiple devices exists'.format(sensor_id))
            if len(resources) == 0:
                raise FalconAPIDataError('Cannot process event for device {}, device not known'.format(sensor_id))
            detail = self.falcon_api.device_details(sensor_id)[0]
            self._host_detail[sensor_id] = detail

        return self._host_detail[sensor_id]

    def azure_arc_config(self, sensor_id):
        if not sensor_id:
            return EventDataError("Cannot fetch Azure Arc info. SensorId field is missing")
        if sensor_id not in self._arc_config:
            is_linux = self.device_details(sensor_id)['platform_name'] == 'Linux'
            path = '/var/opt/azcmagent/agentconfig.json' if is_linux else 'C:\\ProgramData\\AzureConnectedMachineAgent\\Config\\agentconfig.json'
            log.info('Fetching Azure Arc Config %s from the system %s', path, sensor_id)
            file_bytes = self.falcon_api.rtr_fetch_file(sensor_id, path)
            log.info('Fetched Azure Arc Config from the system: %s', str(file_bytes))
            self._arc_config[sensor_id] = json.loads(file_bytes)
        return self._arc_config[sensor_id]

    def mdm_identifier(self, sensor_id, event_platform):
        if not sensor_id:
            return EventDataError("Cannot process event. SensorId field is missing: ")

        if sensor_id not in self._mdm_id or self._mdm_id[sensor_id] is None:
            session = self.falcon_api.init_rtr_session(sensor_id)
            if event_platform == 'Windows':
                command = self.falcon_api.execute_rtr_command(
                    'RTR_ExecuteCommand',
                    session[0]['session_id'],
                    'reg query',
                    'reg query "HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Provisioning\\OMADM\\MDMDeviceID" DeviceClientId'
                )
                response = self.falcon_api.check_rtr_command_status(command[0]['cloud_request_id'], 0)[0]
                while not response['complete']:
                    response = self.falcon_api.check_rtr_command_status(command[0]['cloud_request_id'], 0)[0]
                if response['stderr']:
                    self._mdm_id[sensor_id] = None
                else:
                    self._mdm_id[sensor_id] = response['stdout'].split(' = ')[1].split('\n')[0]
            elif event_platform == 'Mac':
                command = self.falcon_api.execute_rtr_command(
                    'RTR_ExecuteAdminCommand',
                    session[0]['session_id'],
                    'runscript',
                    "runscript -Raw=```system_profiler SPHardwareDataType | awk '/UUID/ { print $3; }'```"
                )
                response = self.falcon_api.check_rtr_command_status(command[0]['cloud_request_id'], 0)[0]
                while not response['complete']:
                    response = self.falcon_api.check_rtr_command_status(command[0]['cloud_request_id'], 0)[0]
                if response['stderr']:
                    self._mdm_id[sensor_id] = None
                else:
                    self._mdm_id[sensor_id] = response['stdout'].split('\n')[0]
            else:
                self._mdm_id[sensor_id] = None

        return self._mdm_id[sensor_id]


class FalconEvent():
    def __init__(self, original_event: Event, cache: FalconCache):
        self.original_event = original_event
        self.cache = cache

    @property
    def device_details(self):
        return self.cache.device_details(self.original_event.sensor_id)

    @property
    def mdm_identifier(self):
        device_details = self.cache.device_details(self.original_event.sensor_id)
        return self.cache.mdm_identifier(self.original_event.sensor_id, device_details['platform_name'])

    def azure_arc_config(self):
        return self.cache.azure_arc_config(self.original_event.sensor_id)

    @property
    def cloud_provider(self):
        return self.device_details.get('service_provider', None)

    @property
    def cloud_provider_account_id(self):
        return self.device_details.get('service_provider_account_id')

    @property
    def instance_id(self):
        return self.device_details.get('instance_id', None)

    @property
    def falcon_link(self):
        return self.original_event['event']['FalconHostLink']

    @property
    def event_id(self):
        return self.original_event['event']['DetectId']

    @property
    def time(self):
        return self.original_event.creation_time

    @property
    def event_create_time(self):
        return self.original_event['metadata']['eventCreationTime']

    @property
    def severity(self):
        return self.original_event['event']['SeverityName']

    @property
    def severity_value(self):
        return self.original_event['event']['Severity']

    @property
    def detect_description(self):
        return self.original_event['event']['DetectDescription']

    @property
    def detect_name(self):
        return self.original_event['event']['DetectName']

    @property
    def service_name(self):
        return self.original_event['event']['ServiceName']

    @property
    def audit_key_values(self):
        return 'AuditKeyValues' in self.original_event['event']



from fig.backends import chronicle
from . import aws
from . import aws_sqs
from . import azure
from . import gcp
from . import workspaceone
from . import cloudtrail_lake
from ..config import config
from ..log import log


ALL_BACKENDS = {
    'AWS': aws,
    'AWS_SQS': aws_sqs,
    'AZURE': azure,
    'GCP': gcp,
    'WORKSPACEONE': workspaceone,
    'CHRONICLE': chronicle,
    'CLOUDTRAIL_LAKE': cloudtrail_lake
}


class Backends():
    def __init__(self):
        self.runtimes = [v.Runtime()
                         for k, v in ALL_BACKENDS.items()
                         if k in config.backends]
        if len(self.runtimes) == 0:
            raise Exception("No Backend enabled. Exiting.")

        accepted_types = self.relevant_event_types
        if accepted_types is None:
            log.info("At least one of the enabled backends will receive all the events")
        else:
            log.info("Enabled backends will only process events with types: %s", accepted_types)

    def process(self, falcon_event):
        for runtime in self.runtimes:
            if self.event_type_is_accepted(runtime, falcon_event) and self.cloud_detection_is_relevant(falcon_event) and runtime.is_relevant(falcon_event):
                runtime.process(falcon_event)

    def cloud_detection_is_relevant(self, falcon_event):
        if falcon_event.original_event.event_type == 'DetectionSummaryEvent':
            if falcon_event.cloud_provider in config.detections_exclude_clouds or falcon_event.cloud_provider is None and 'unrecognized' in config.detections_exclude_clouds:
                log.debug('A detection event is skipped based on cloud based exclusion')
                return False
        return True

    def event_type_is_accepted(self, runtime, falcon_event):
        return runtime.RELEVANT_EVENT_TYPES == "ALL" or falcon_event.original_event.event_type in runtime.RELEVANT_EVENT_TYPES

    @property
    def relevant_event_types(self):
        if any("ALL" in r.RELEVANT_EVENT_TYPES for r in self.runtimes):
            return None
        return set(typ for r in self.runtimes for typ in r.RELEVANT_EVENT_TYPES)



import json
from functools import lru_cache
import boto3
from botocore.exceptions import ClientError
from ...config import config
from ...log import log


class Submitter():
    def __init__(self):
        aws_region = config.get('aws_sqs', 'region')
        log.debug("Attempting to connect to AWS (region %s) SQS: %s", aws_region, self.sqs_queue_name)
        try:
            aws_client = boto3.resource('sqs', region_name=aws_region)
            self.queue = aws_client.get_queue_by_name(QueueName=self.sqs_queue_name)
        except ClientError:  # pylint: disable=W0703
            log.exception("Cannot configure AWS SQS Queue: %s in %s", aws_region, self.sqs_queue_name)
            self.queue = None
            raise

    def submit(self, event):
        eoe = event.original_event
        if self.is_fifo:
            feed_id = getattr(eoe, 'feed_id') if hasattr(eoe, 'feed_id') else 0
            self.queue.send_message(
                MessageGroupId="fig/%s/%s" % (self.app_id, feed_id),
                MessageDeduplicationId=str(eoe.offset),
                MessageBody=json.dumps(eoe)
            )
        else:
            self.queue.send_message(MessageBody=json.dumps(eoe))

    @property
    @lru_cache
    def is_fifo(self):
        return self.sqs_queue_name.endswith('.fifo')

    @property
    @lru_cache
    def sqs_queue_name(self):
        return config.get('aws_sqs', 'sqs_queue_name')

    @property
    @lru_cache
    def app_id(self):
        return config.get('falcon', 'application_id')


class Runtime():
    RELEVANT_EVENT_TYPES = "ALL"

    def __init__(self):
        log.info("AWS SQS Backend is enabled.")
        self.submitter = Submitter()

    def is_relevant(self, falcon_event):
        return True

    def process(self, falcon_event):
        self.submitter.submit(falcon_event)


__all__ = ['Runtime']



from json import dumps
import boto3
from botocore.exceptions import ClientError
from ...config import config
from ...log import log
from .cloudtrail_offset import LastEventOffset


class Submitter():
    def __init__(self, event, channel_arn, region, account_id, last_event_offset):
        self.event = event
        self.channel_arn = channel_arn
        self.region = region
        self.account_id = account_id
        self.last_event_offset = last_event_offset

    def cloudtrail_lake_audit_event(self):
        '''
        Map the audit event and return a dict of the data.
        '''
        uid = self.event.original_event.uid
        event = self.event.original_event['event']
        metadata = self.event.original_event['metadata']
        event_time = self.event.time.isoformat(timespec='seconds') + 'Z'

        user_identity = {
            "type": metadata['eventType'],
            "principalId": event['UserId'],
            "details": {"AuditKeyValues": "n/a"}
        }

        # Not all events have AuditKeyValues
        if self.event.audit_key_values:
            user_identity['details']['AuditKeyValues'] = event['AuditKeyValues']

        event_data = {
            "version": metadata['version'],
            "userIdentity": user_identity,
            "userAgent": "falcon-integration-gateway",
            "eventSource": "crowdstrike",
            "eventName": event['OperationName'],
            "eventTime": event_time,
            "UID": uid,
            "sourceIPAddress": event['UserIp'],
            "recipientAccountId": self.account_id,
            "additionalEventData": {
                "raw": self.event.original_event
            }
        }

        return event_data

    def send_to_cloudtraillake(self, event_data):
        '''
        Sends the event to CloudTrail Lake. Returns the response.
        '''
        client = boto3.client('cloudtrail-data', region_name=self.region)
        response = False
        try:
            response = client.put_audit_events(
                auditEvents=[
                    {
                        'id': event_data['UID'],
                        'eventData': dumps(event_data)
                    }
                ],
                channelArn=self.channel_arn
            )
        except ClientError as err:
            log.exception(str(err))

        return response

    def submit(self):
        '''
        Submit events to CloudTrail Lake and updates the last seen offset.
        '''
        operation_name = self.event.original_event['event']['OperationName']
        uid = self.event.original_event.uid
        log.info("Processing user activity event: %s ID: %s", operation_name, uid)

        event_data = self.cloudtrail_lake_audit_event()
        response = self.send_to_cloudtraillake(event_data)

        # Check response for errors
        if not response:
            log.error("Exception Error occured while sending event to CloudTrail Lake.")
        else:
            if response['failed']:
                log.error("Failed Response recieved for: %s", response['failed'])
            else:
                log.info("Successfully sent event ID: %s to CloudTrail Lake. (Request ID: %s)",
                         uid, response['ResponseMetadata']['RequestId'])
                # Update the last seen offset for this feed
                self.last_event_offset.update_last_seen_offsets(self.event.original_event.feed_id,
                                                                self.event.original_event.offset)


class Runtime():
    RELEVANT_EVENT_TYPES = ['AuthActivityAuditEvent']

    def __init__(self):
        log.info("AWS CloudTrail Lake Backend is enabled.")
        # Initialize Config Variables
        self.channel_arn = config.get('cloudtrail_lake', 'channel_arn')
        self.region = config.get('cloudtrail_lake', 'region')
        self.account_id = None
        try:
            self.account_id = self.channel_arn.split(':')[4]
        except IndexError:
            log.error("Could not get account ID from Channel ARN: %s", self.channel_arn)
            raise
        # Instantiate the last seen offset object
        self.last_event_offset = LastEventOffset()
        # Get the last seen offset for each feed
        self.last_seen_offsets = self.last_event_offset.get_last_seen_offsets()

    def is_relevant(self, falcon_event):
        if falcon_event.service_name.lower() != "crowdstrike authentication":
            return False

        feed_id = falcon_event.original_event.feed_id
        offset = falcon_event.original_event.offset
        if offset > self.last_seen_offsets.get(feed_id, 0):
            return True
        return False

    def process(self, falcon_event):
        Submitter(falcon_event, self.channel_arn, self.region, self.account_id, self.last_event_offset).submit()


__all__ = ['Runtime']
